"""Functions for reading and interpreting the DISPLACEMENTS file."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'

import logging
import re

import numpy as np

from viperleed.calc.lib.base import readIntRange
from viperleed.calc.lib.base import splitSublists
from viperleed.calc.symmetry import enforceSymmetry
from viperleed.calc.symmetry import setSymmetry


logger = logging.getLogger(__name__)


def readDISPLACEMENTS(rp, filename="DISPLACEMENTS"):
    """
    Reads the DISPLACEMENTS file to rp.disp_blocks without interpreting it.

    Parameters
    ----------
    rp : Rparams
        The run parameters object.
    filename : str, optional
        The file name to read from. The default is "DISPLACEMENTS".

    Raises
    ------
    RuntimeError
        Raised if unexpected syntax is encountered and cannot be resolved.

    Returns
    -------
    None.

    """
    rp.disp_blocks = []
    copyblock = None
    loopStarts = []
    loopLine = False
    try:
        with open(filename, 'r') as rf:
            lines = rf.readlines()
    except FileNotFoundError:
        logger.error("DISPLACEMENTS not found.")
        raise
    except Exception:
        logger.error("Error reading DISPLACEMENTS file.")
        raise
    for (i, line) in enumerate(lines):
        for c in ["!", "#", "%"]:    # start of comment
            line = line.split(c)[0].strip()
        if re.search(r'<\s*/?\s*loop\s*>', line.lower()):
            if "=" in line:
                logger.warning("DISPLACEMENTS file: One line cannot contain "
                               "both an '=' sign and a <loop> flag.")
                rp.setHaltingLevel(2)
                continue
            loopLine = True
            ltags = re.findall(r'<\s*/?\s*loop\s*>', line.lower())
            n_blocks_real = len(rp.disp_blocks)
            if copyblock:
                n_blocks_real += 1
            for t in ltags:
                if re.match(r'<\s*loop\s*>', t):
                    loopStarts.append(n_blocks_real)
                else:
                    if len(loopStarts) == 0:
                        logger.warning("DISPLACEMENTS file: Unmatched "
                                       "</loop> flag found, skipping.")
                        rp.setHaltingLevel(2)
                        continue
                    if loopStarts[-1] == n_blocks_real:
                        logger.warning("DISPLACEMENTS file: Found empty "
                                       "loop.")
                        loopStarts.pop()
                        rp.setHaltingLevel(1)
                        continue
                    rp.disp_loops.append((loopStarts.pop(), n_blocks_real - 1))
            continue
        if '=' not in line:
            continue
        if line.startswith('=='):
            if re.match(r'==\s*s', line.lower()):  # search block
                loopLine = False
                try:
                    name = (line[re.match(r'==\s*s(earch)?\s+',
                                          line.lower()).span()[1]:].strip())
                except Exception:
                    name = ""
                names = [n for (_, n) in rp.disp_blocks]
                names.extend([n[:-6] for (_, n) in rp.disp_blocks
                              if n.endswith(" [0 1]") or n.endswith(" [1 0]")])
                if not name or name in names:  # get unique name
                    if not name:
                        i = 1
                    else:
                        i = 2
                    nn = (name + " {}".format(i)).strip()
                    while nn in names:
                        i += 1
                        nn = (name + " {}".format(i)).strip()
                    name = nn
                if not rp.disp_blocks or len(rp.disp_blocks[-1][0]) != 0:
                    if copyblock:
                        rp.disp_blocks.append(copyblock)
                        copyblock = None
                    rp.disp_blocks.append(([], name))
                else:
                    rp.disp_blocks[-1] = ([], name)
                continue
            elif line.split('==')[1].strip().lower()[0] != "d":
                logger.warning("Found DISPLACEMENTS line starting with '==',"
                               "but could not interpret it as the start of a "
                               "search or domain block.")
                rp.setHaltingLevel(2)
                continue
        elif loopLine and not len(rp.disp_blocks) == 0:
            logger.error("DISPLACEMENTS file line {}: Line with <loop> or "
                         "</loop> flags must be followed by a new search "
                         "block or end of file!".format(i+1))
            raise RuntimeError("Misplaced <loop> flag in DISPLACEMENTS file.")
        loopLine = False
        if len(rp.disp_blocks) == 0:
            rp.disp_blocks.append(([], ""))
        p = line.split("=")[0]
        if ("xy" in p or "ab" in p) and "[" not in p:  # shorthand for 2 blocks
            if not copyblock:
                if rp.disp_blocks[-1][1]:
                    names = (rp.disp_blocks[-1][1]+" [1 0]",
                             rp.disp_blocks[-1][1]+" [0 1]")
                else:
                    names = ("", "")
                rp.disp_blocks[-1] = (rp.disp_blocks[-1][0], names[0])
                copyblock = (rp.disp_blocks[-1][0][:], names[1])
            if "xy" in p:
                s = "xy"
            else:
                s = "ab"
            copyblock[0].append(line.replace(s, s+"[0 1]", 1))
            line = line.replace(s, s+"[1 0]", 1)

        elif copyblock:
            copyblock[0].append(line)
        rp.disp_blocks[-1][0].append(line)
    if copyblock:
        rp.disp_blocks.append(copyblock)
    if len(rp.disp_blocks[-1][0]) == 0 and len(rp.disp_blocks) > 1:
        rp.disp_blocks = rp.disp_blocks[:-1]
    if len(loopStarts) != 0:
        logger.warning("DISPLACEMENTS file: Unmatched <loop> flags found, "
                       "loops are still open at end of file.")
        rp.setHaltingLevel(2)
    if not rp.domainParams:
        return
    # in case of domains, now split blocks to domains
    for dp in rp.domainParams:
        dp.rp.disp_blocks = []
    for (lines, blockname) in rp.disp_blocks:
        d = ""
        dlines = {}
        dnum = 1
        names = [dp.name for dp in rp.domainParams]
        skipblock = False
        for name in names:
            dlines[name] = []
        for line in lines:
            if '=' not in line:
                continue
            if re.match(r'==\s*d', line.lower()):  # start domain block
                skipblock = False
                try:
                    d = (line[re.match(r'==\s*d(omain)?\s+',
                                       line.lower()).span()[1]:].strip())
                except Exception:
                    if str(dnum) in names:
                        d = str(dnum)
                        dnum += 1
                    else:
                        logger.warning(
                            "DISPLACEMENTS file: found start of domain block "
                            "with missing domain name. Block will be skipped.")
                        rp.setHaltingLevel(2)
                        d = ""
                        skipblock = True
                if d and d not in names:
                    logger.warning(
                        "DISPLACEMENTS file: found start of domain "
                        "block with invalid domain name {}. Block will be "
                        "skipped.".format(d))
                    d = ""
                    skipblock = True
            elif skipblock:
                continue
            else:
                if not d:
                    logger.warning(
                        "DISPLACEMENTS file: found line outside "
                        "of a domain block, line will be skipped: "+line)
                else:
                    dlines[d].append(line)
        for dp in rp.domainParams:
            dp.rp.disp_blocks.append((dlines[dp.name], blockname))
    return


def readDISPLACEMENTS_block(rp, sl, dispblock, only_mode=""):
    """
    Reads a block from the DISPLACEMENTS file and adds the information to
    all atoms in the slab.

    Parameters
    ----------
    rp : Rparams
        The run parameters object.
    sl : Slab
        The slab object containing atom and site information, to which the
        displacements will be written.
    dispblock : tuple (lines, name)
        The information in the DISPLACEMENTS block to be interpreted, as read
        be readDISPLACEMENTS.
    only_mode : str, optional
        If set to 'vib', 'geo' or 'occ', all other displacements
        will be skipped. This is meant to be used for error calculations, where
        only one-dimensional deltas are desired.

    Raises
    ------
    RuntimeError
        Raised for invalid or inconsistent input.
    ValueError
        Raised if displacement lists have inconsistent length, or if
        site occupation exceeds 1.

    Returns
    -------
    deltas_required : bool
        True if any displacements were assigned, False otherwise

    """
    deltas_required = False
    abst = sl.ab_cell.T
    # if the unit cell gets modified by SYM_DELTA, restore it afterwards
    uCellState = sl.ucell_mod
    (lines, name) = dispblock
    mode = 0  # 0: not reading; 1: reading geo, 2: reading vib, 3: reading occ
    regex = False   # read regular expressions as-is or not
    constraints = []  # collect constraints here first, interpret after loop
    for line in lines:
        if "!" in line:
            line = line.split("!")[0]
        line = line.strip()
        if '=' in line:
            if line[0] == '=':
                llist = line[1:].split()
                if llist[0][0].lower() == 'g':
                    mode = 1  # geometry
                    if only_mode and only_mode != "geo":
                        mode = 0
                elif llist[0][0].lower() == 'v':
                    mode = 2  # vibration
                    if only_mode and only_mode != "vib":
                        mode = 0
                elif llist[0][0].lower() == 'o':
                    mode = 3  # occupation
                    if only_mode and only_mode != "occ":
                        mode = 0
                elif llist[0][0].lower() == 'c':
                    mode = 4  # constraint
                elif llist[0][0].lower() == 'r':
                    regex = True
                    if len(llist) >= 2:
                        if llist[1].lower() == 'off':
                            regex = False
                else:
                    logger.warning("DISPLACEMENTS: Found line starting with "
                                   "'=', but didn't recognize flag.")
                    rp.setHaltingLevel(1)
                continue
            elif mode == 0:
                continue
            else:
                pside = line.split('=')[0].strip()
                if pside:
                    # get rid of spaces and check the leftmost entry.
                    plist = pside.split()
                    if plist:
                        param = plist[0]
                        if param[0] == '!':
                            continue
                else:
                    continue
        else:
            continue
        value = line.split('=')[1].strip()
        try:
            llist = value.split()
        except IndexError:
            logger.warning("DISPLACEMENTS file: " + param + " appears to "
                           "have no value")
            rp.setHaltingLevel(1)
            continue
        if not llist:
            logger.warning("DISPLACEMENTS file: " + param + " appears to "
                           "have no value")
            rp.setHaltingLevel(1)
            continue
        if param.lower() == "sym_delta":
            s = llist[0].lower()
            grouplist = [
                "p1", "p2", "pm", "pg", "cm", "rcm", "pmm", "pmg", "pgg",
                "cmm", "rcmm", "p4", "p4m", "p4g", "p3", "p3m1", "p31m",
                "p6", "p6m"]  # TODO: use guilib or put as const in leedbase
            targetsym = ""
            if s[0] == "t":
                # True - go to highest symmetry
                targetsym = "found"
            elif s[0] == "f":
                # False - set p1
                targetsym = "p1"
            elif s in grouplist:
                if s not in ["cm", "pmg"]:
                    targetsym = s
                else:
                    logger.warning('DISPLACEMENTS file: SYM_DELTA: For '
                                   'group '+s+', direction needs to be '
                                   'specified. Input will be ignored.')
                    rp.setHaltingLevel(2)
            elif s[:2] in ["pm", "pg", "cm"] or s[:3] == "pmg":
                # regex to read
                rgx = re.compile(
                    r'\s*(?P<group>(pm|pg|cm|rcm|pmg))\s*'
                    r'\[\s*(?P<i1>[-012]+)\s+(?P<i2>[-012]+)\s*\]')
                m = rgx.match(value.lower())
                if not m:
                    logger.warning(
                        'DISPLACEMENTS file: SYM_DELTA: Could not parse given '
                        'value. Input will be ignored.')
                    rp.setHaltingLevel(2)
                else:
                    i1 = i2 = -2
                    group = m.group('group')
                    try:
                        i1 = int(m.group('i1'))
                        i2 = int(m.group('i2'))
                    except ValueError:
                        logger.warning('DISPLACEMENTS file: SYM_DELTA: '
                                       'Could not parse given value. '
                                       'Input will be ignored.')
                        rp.setHaltingLevel(1)
                    if (group in ["pm", "pg", "cm", "rcm", "pmg"]
                            and i1 in range(-1, 3) and i2 in range(-1, 3)):
                        targetsym = m.group(0)
                    else:
                        logger.warning('DISPLACEMENTS file: SYM_DELTA: '
                                       'Could not parse given value. '
                                       'Input will be ignored.')
                        rp.setHaltingLevel(2)
            else:
                logger.warning('DISPLACEMENTS file: SYM_DELTA: Could not '
                               'parse given value. Input will be ignored.')
                rp.setHaltingLevel(2)
            if targetsym != "":
                sl.revert_unit_cell(uCellState)
                setSymmetry(sl, rp, targetsym)
                enforceSymmetry(sl, rp, movement=False, rotcell=False)
            continue
        # if we are still here, then it's a "normal" line
        ats = []
        if mode == 1:
            rgx = re.compile(
                r'\s*(?P<nums>(([-:0-9]|((l|L)\([-:0-9\s]+\))'
                r')\s*)*)(?P<dir>(z|((ab|xy)\s*)?\[\s*[-0-9]+\s+[-0-9]+\s*\]|'
                r'(azi|r)\(?((ab|xy)\s*)?\[\s*[-0-9\.]+\s+[-0-9\.]+\s*\]\)?)?'
                r'(\s*offset)?)')
        else:
            rgx = re.compile(r'(?P<nums>(([-:0-9]|((l|L)\([-:0-9\s]+\)))\s*)*)'
                             r'(\s*offset)?')
        if mode == 4:
            if param.lower() == "geo":
                ctype = 1
            elif param.lower() == "vib":
                ctype = 2
            elif param.lower() == "occ":
                ctype = 3
            else:
                logger.warning("DISPLACEMENTS file: CONSTRAINT flag " + param +
                               " not recognized. Input will be ignored.")
                rp.setHaltingLevel(2)
                continue
            pspec = pside.split(maxsplit=1)[1]
        else:
            pspec = pside
        _break = True
        for s in pspec.split(","):
            s = s.strip()
            sname = s.split()[0]
            spec = s.split(sname, 1)[1].strip().lower()
            m = rgx.match(spec)
            if not m or (mode == 1 and not m.group("dir")):
                logger.warning('DISPLACEMENTS file: could not parse as '
                               'numbers or direction, skipping line: '+pside)
                rp.setHaltingLevel(2)
                break
            ats.append((sname, m.group("nums")))
            if mode == 1:
                dr = m.group("dir")
                if not dr:
                    logger.warning('DISPLACEMENTS file: could not parse '
                                   'direction, skipping line: '+pside)
                    rp.setHaltingLevel(2)
                    break
                # quick consistency check on direction:
                if "[" in dr:
                    if "azi" not in dr and "r" not in dr:
                        try:
                            dirints = (
                                int(dr.split("[")[1].strip().split()[0]),
                                int(dr.split("]")[0].strip().split()[-1]))
                        except ValueError:
                            logger.warning(
                                'DISPLACEMENTS file: could not parse '
                                'direction "'+dr+'", skipping line: '+pside)
                            rp.setHaltingLevel(2)
                            break
                    else:
                        try:
                            dirfloats = (
                                float(dr.split("[")[1].strip().split()[0]),
                                float(dr.split("]")[0].strip().split()[-1]))
                        except ValueError:
                            logger.warning(
                                'DISPLACEMENTS file: could not parse '
                                'direction "'+dr+'", skipping line: '+pside)
                            rp.setHaltingLevel(2)
                            break
        else:
            _break = False
        if _break:
            rp.setHaltingLevel(2)
            continue
        # get list of atoms to manipulate
        targetAtEls = []
        _break = True
        for (sname, nums) in ats:
            numlist = nums.split()
            if numlist:
                # get proper numlist as integers
                # first need to recombine expressions like "L(2 3)"
                nl = []
                partstring = ""
                bracketcount = 0
                for s in numlist:
                    partstring += s
                    bracketcount += s.count("(") - s.count(")")
                    if bracketcount == 0:
                        nl.append(partstring)
                        partstring = ""
                    else:
                        partstring += " "
                if bracketcount != 0:
                    logger.warning("DISPLACEMENTS file: Bracket mismatch in "
                                   "line " + pside)
                    if len(partstring) > 0:
                        nl.append(partstring)
                numlist = []
                for s in nl:
                    if "l(" not in s:
                        il = readIntRange(s)
                        if len(il) > 0:
                            numlist.extend(il)
                        else:
                            logger.warning(
                                'DISPLACEMENTS file: could not '
                                'parse integer range, skipping line: '+pside)
                            rp.setHaltingLevel(2)
                            break
                    else:
                        m = re.match(r'l\((?P<laynum>[-:0-9\s]+)\)', s)
                        if not m:
                            logger.warning(
                                'DISPLACEMENTS file: could not parse layer '
                                'expression, skipping line: '+pside)
                            rp.setHaltingLevel(2)
                            break
                        il = readIntRange(m.group("laynum"))
                        if len(il) == 0:
                            logger.warning(
                                'DISPLACEMENTS file: could not parse layer '
                                'expression, skipping line: '+pside)
                            rp.setHaltingLevel(2)
                            break
                        for ln in il:
                            if ln > sl.n_layers:
                                logger.warning(
                                    'DISPLACEMENTS file: layer number out of '
                                    'bounds, skipping line: '+pside)
                                rp.setHaltingLevel(2)
                                break
                            numlist.extend([at.num for at in sl
                                            if at.layer == sl.layers[ln-1]])
                else:  # loop finished without beak
                    _break = False
                if _break:
                    break
                _break = True
            # interpret label:
            targetsites = []
            # first try with POSCAR element:
            for site in sl.sitelist:
                s = sname
                if "_" not in s:
                    if regex:
                        s += ".*"
                    else:
                        s += "*"
                if not regex:
                    s = re.escape(s)   # double-slash non-literal characters
                    # if regular expressions are not enabled, we want to
                    #  still interpret * as "any number of any characters"
                    s = s.replace('\\*', '.*')
                m = re.match(s, site.label)
                if m:
                    # if the length of the matched text == the site label,
                    #  it's a real match
                    if m.end(0) == len(site.label):
                        targetsites.append(site)
            if len(targetsites) > 0:
                targetel = ""   # use ALL elements of site
            elif mode == 3:
                # if we're reading occupation, only POSCAR elements are
                #   allowed -> error
                logger.warning(
                    'DISPLACEMENTS file: could not parse '+sname+' as POSCAR '
                    'element or site label, skipping line.')
                rp.setHaltingLevel(2)
                break
            else:
                # check ELEMENT_MIX elements:
                el = sname.split("_")[0].capitalize()
                s = el
                if not regex:
                    s = re.escape(s)   # double-slash
                    s = s.replace('\\*', '.*')
                cels = [cel for cel in sl.chemelem if re.match(s, cel)]
                if el not in sl.elements and len(cels) > 0:
                    pel = ""
                    for k in rp.ELEMENT_MIX:
                        if any(e in rp.ELEMENT_MIX[k] for e in cels):
                            pel = k
                            targetel = [e for e in cels
                                        if e in rp.ELEMENT_MIX[k]][0]
                            break
                    s = sname.replace(sname.split("_")[0], pel, 1)
                    if s in sl.elements:  # is element, all sites
                        targetsites.extend([site for site in sl.sitelist if
                                            site.el == s])
                    else:
                        for site in sl.sitelist:
                            if not regex:
                                s = re.escape(s)   # double-slash
                                s = s.replace('\\*', '.*')
                            m = re.match(s, site.label)
                            if m:
                                if m.end(0) == len(site.label):
                                    targetsites.append(site)
                else:
                    logger.warning(
                        'DISPLACEMENTS file: could not parse '+sname+' as '
                        'element or site label, skipping line.')
                    rp.setHaltingLevel(2)
                    break
                if len(targetsites) == 0:
                    logger.warning(
                        'DISPLACEMENTS file: could not parse '+sname+' as '
                        'element or site label, skipping line.')
                    rp.setHaltingLevel(2)
                    break
            for at in sl:
                if ((at.num in numlist or len(numlist) == 0)
                        and at.site in targetsites and not at.is_bulk):
                    targetAtEls.append((at, targetel))
        else:  # loop finished without break
            _break = False
        if _break:
            continue  # error message has already happened
        if len(targetAtEls) == 0:
            logger.warning('DISPLACEMENTS file: no atoms found for '
                           'line, line will be skipped: '+pside)
            rp.setHaltingLevel(1)
            continue
        if mode == 2 or (mode == 1 and "offset" not in dr):
            # geometrical or vibrational displacement, get range:
            try:
                fl = [float(s) for s in llist]
            except ValueError:
                if ("offset" in pside and llist[0].lower() in ["clear",
                                                               "original"]):
                    fl = [0.]
                elif "offset" in pside and llist[0].lower().startswith("cont"):
                    continue
                    # instruction to use value from previous search
                    # -> default, ignore
                else:
                    fl = []
                    logger.warning(
                        'DISPLACEMENTS file: could not parse ' + value +
                        ' as list of floats, skipping line.')
                    rp.setHaltingLevel(1)
                    continue
            if len(fl) < 3:
                if mode == 2 and len(fl) == 1:  # interpret as static offset
                    fl = [fl[0], fl[0], 1]
                else:
                    logger.warning('DISPLACEMENTS file: too few values '
                                   'found, skipping line: '+pside)
                    rp.setHaltingLevel(1)
                    continue
            #  OLD: force uneven number of steps
            # steps = abs(int(round((fl[1]-fl[0]) / fl[2])))+1
            # mid = (fl[1]+fl[0]) / 2
            # if steps % 2 == 0:      # even number of steps, extend range
            #     steps += 1
            # drange = np.arange(mid - ((steps-1)/2*fl[2]),
            #                    mid + ((steps-1)/2*fl[2])+1e-6, fl[2])
            if fl[0] < fl[1]:
                drange = np.arange(fl[0], fl[1] + 1e-6, abs(fl[2]))
            else:
                drange = np.arange(fl[0], fl[1] - 1e-6, -abs(fl[2]))
            if min([abs(v) for v in drange]) > 5e-5:
                logger.warning(
                    "DISPLACEMENTS: A range does not contain zero. This means "
                    "that keeping the input configuration unchanged will not "
                    "be possible during the search: " + value)
                rp.setHaltingLevel(1)
        if mode == 1 and "offset" in dr:
            # geo offset, get value
            try:
                offval = float(llist[0])
            except (ValueError, IndexError):
                if llist[0].lower() in ["clear", "original"]:
                    offval = 0.
                elif llist[0].lower().startswith("cont"):
                    continue
                else:
                    logger.warning('DISPLACEMENTS file: could not parse '
                                   + value + ' as float, skipping line.')
                    rp.setHaltingLevel(1)
                    continue
            drange = [offval]
        if mode == 1:
            if dr.strip() == "offset":
                # No direction
                if offval != 0.:
                    logger.warning('DISPLACEMENTS file: cannot assign '
                                   'geo offset : no direction given: '+pside)
                    rp.setHaltingLevel(1)
                    continue
                else:
                    for (at, targetel) in targetAtEls:
                        at.clearOffset(1, targetel)
                    continue
            # geometrical displacement, get direction vector:
            drvec = np.array([0., 0., 0.])
            if "azi" not in dr and "r" not in dr:
                if "z" in dr:
                    drvec[2] = -1.  # invert dir: positive = away from bulk
                elif "xy" in dr:
                    v = np.array(dirints)
                    v = v / np.linalg.norm(v)
                    drvec[:2] = v
                else:
                    v = dirints[0]*abst[0] + dirints[1]*abst[1]
                    v = v / np.linalg.norm(v)
                    drvec[:2] = v
                disprange = [f*drvec for f in drange]
                for (at, targetel) in targetAtEls:
                    # check whether displacement is allowed
                    if "z" in dr:
                        allowed = True
                    elif type(at.freedir) == int:
                        if at.freedir == 0:
                            allowed = False
                        else:
                            allowed = True
                    else:
                        freev = at.freedir[0]*abst[0] + at.freedir[1]*abst[1]
                        freev = freev / np.linalg.norm(freev)
                        if abs(abs(np.dot(drvec[:2], freev))-1) < 1e-5:
                            allowed = True
                        else:
                            allowed = False
                    if allowed:
                        if "offset" not in dr:
                            at.assignDisp(mode, disprange, targetel,
                                          disp_label=dr,
                                          disp_lin_steps=drange)
                        else:
                            at.clearOffset(1, targetel)
                            at.assignDisp(4, disprange, targetel)
                        deltas_required = True
                    else:
                        logger.warning(
                            f'In-plane displacement assignment for {at} is '
                            'forbidden by symmetry and will be skipped. See '
                            '"FreeDir" in POSCAR file. To apply this '
                            'displacement, use either the SYMMETRY_FIX '
                            'parameter to lower the symmetry, or use '
                            'SYM_DELTA in the DISPLACEMENTS file to allow '
                            'symmetry breaking for this atom.'
                            )
            else:
                if "xy" in dr:
                    c = np.array(dirfloats)
                else:
                    c = dirfloats[0]*abst[0] + dirfloats[1]*abst[1]
                for (at, targetel) in targetAtEls:
                    if "r" in dr:
                        v = at.cartpos[:2] - c
                        v = v / np.linalg.norm(v)
                        allowed = True
                        if type(at.freedir) == int and at.freedir == 0:
                            allowed = False
                        elif type(at.freedir) != int:
                            freev = (at.freedir[0]*abst[0] +
                                     at.freedir[1]*abst[1])
                            freev = freev / np.linalg.norm(freev)
                            if abs(abs(np.dot(v, freev))-1) > 1e-5:
                                allowed = False
                        if allowed:
                            drvec = np.array([0., 0., 0.])
                            drvec[:2] = v
                            disprange = [f*drvec for f in drange]
                            if "offset" not in dr:
                                at.assignDisp(mode, disprange, targetel,
                                              disp_label=dr,
                                              disp_lin_steps=drange)
                            else:
                                at.assignDisp(4, disprange, targetel)
                            deltas_required = True
                    else:  # azi
                        # allowed only for completely free atoms
                        if type(at.freedir) == int and at.freedir == 1:
                            r = at.cartpos[:2] - c
                            disprange = []
                            for d in drange:
                                # disprange based on rotating vector around c
                                ar = d / np.linalg.norm(r)  # rotation angle
                                tm = np.array([[np.cos(ar)-1, -np.sin(ar)],
                                               [np.sin(ar), np.cos(ar)-1]])
                                disprange.append(np.append(np.dot(tm, r), 0.))
                            if "offset" not in dr:
                                at.assignDisp(mode, disprange, targetel,
                                              disp_label=dr,
                                              disp_lin_steps=drange)
                            else:
                                at.clearOffset(1, targetel)
                                at.assignDisp(4, disprange, targetel)
                            deltas_required = True
                        else:
                            logger.warning(
                                'In-plane azimuthal displacement assignment '
                                f'for {at} is forbidden by symmetry and will '
                                'be skipped. See "FreeDir" in POSCAR file. '
                                'To apply this displacement, use either the '
                                'SYMMETRY_FIX parameter to lower the symmetry,'
                                ' or use SYM_DELTA in the DISPLACEMENTS file '
                                'to allow symmetry breaking for this atom.'
                                )
        elif mode == 2:
            # vibrational displacement, apply:
            for (at, targetel) in targetAtEls:
                if "offset" in pside:
                    at.clearOffset(mode, targetel)
                if not (len(drange) == 1 and drange[0] == 0.):
                    at.assignDisp(mode, drange, targetel,
                                  disp_label="N/A",
                                  disp_lin_steps=drange)
                deltas_required = True
        elif mode == 3:
            # occupations, get ranges:
            sublists = splitSublists(llist, ',')
            maxsteps = 0
            _break = True
            for subl in sublists:
                # get range
                try:
                    fl = [float(s) for s in subl[-3:]]
                except (ValueError, IndexError):
                    fl = []
                    if len(subl) == 2:
                        try:
                            fl = [float(subl[-1])]
                        except Exception:
                            pass
                    if len(fl) == 1 or (len(fl) == 3 and fl[2] < 5e-5):
                        fl = [fl[0], fl[0], 1]
                    if len(fl) < 3:
                        logger.warning(
                            'DISPLACEMENTS file: could not parse ' + value
                            + ' as list of floats, skipping line.')
                        rp.setHaltingLevel(1)
                        break
                if len(fl) < 3 or (len(fl) == 3 and fl[2] < 5e-5):
                    if len(fl) == 1 or (len(fl) == 3 and fl[2] < 5e-5):
                        # interpret as static offset
                        fl = [fl[0], fl[0], 1]
                    else:
                        logger.warning('DISPLACEMENTS file: too few values '
                                       'found, skipping line: '+pside)
                        rp.setHaltingLevel(1)
                        break
                steps = abs(int(round((fl[1]-fl[0]) / fl[2])))+1
                if steps > maxsteps:
                    maxsteps = steps
            else:
                _break = False
            if _break:
                continue
            for subl in sublists:
                try:
                    fl = [float(s) for s in subl[-3:]]
                except (ValueError, IndexError):
                    f = float(subl[-1])  # checked before, will work
                    fl = [f, f, 1]
                if fl[2] < 5e-5:
                    fl = [fl[0], fl[0], 1]
                steps = abs(int(round((fl[1]-fl[0]) / fl[2])))+1
                if steps > maxsteps:
                    steps = maxsteps
                    logger.warning(
                        'DISPLACEMENTS file: inconsistent step '
                        'numbers for occupancies of '+pside+'. Decreasing '
                        'spacing for '+subl[0]+' to make number of steps '
                        'equal.')
                if steps == 1 and maxsteps > 1:
                    drange = np.array([fl[0]] * maxsteps)
                else:
                    mid = (fl[1]+fl[0]) / 2
                    drange = np.arange(mid-((steps-1)/2*fl[2]),
                                       mid+((steps-1)/2*fl[2])+1e-5, fl[2])
                    if fl[1] < fl[0]:
                        drange = np.flip(drange)
                if subl[0].capitalize() in sl.chemelem:
                    targetel = subl[0]
                else:
                    logger.error('DISPLACEMENTS file: '+subl[0]+' not '         # TODO: This should probably raise an exception
                                   'found in element list. No assignment '
                                   'will be made.')
                    rp.setHaltingLevel(2)
                    continue
                for (at, _) in targetAtEls:
                    if targetel in at.disp_occ:
                        at.assignDisp(mode, drange, targetel,
                                      disp_label="N/A",
                                      disp_lin_steps=None)  # for occupation, we figure it out in r_error
                        deltas_required = True
                    else:
                        logger.warning(
                            'DISPLACEMENTS file: trying to address {} with '
                            'wrong element. Atom will be skipped.'.format(at))
                        rp.setHaltingLevel(1)
                # else:
                #     # !!! currently unused, not in wiki - delete?
                #     elparts = "".join(subl[:-3]).split("+")
                #     elweights = []
                #     elsum = 0.0
                #     rgx = re.compile(r'\s*(?P<number>[0-9\.]+)\s*'
                #                      +r'(?P<name>[a-zA-Z]+)')
                #     for part in elparts:
                #         m = rgx.match(part)
                #         if not m:
                #             logger.warning('DISPLACEMENTS file: could '
                #                 'not parse '+part+' as number and '
                #                 'element. No assignment will be made.')
                #             rp.setHaltingLevel(1)
                #             break
                #         try:
                #             f = float(m.group("number"))
                #         except ValueError:
                #             logger.warning('DISPLACEMENTS file: '
                #              'could not parse '+m.group("number")
                #              +' as float. No assignment will be made.')
                #             rp.setHaltingLevel(1)
                #             break
                #         el = m.group("name").capitalize()
                #         if not el in sl.chemelem:
                #             logger.warning('DISPLACEMENTS file: '+el
                #                 +' not found in element list. No '
                #                 'assignment will be made.')
                #             rp.setHaltingLevel(1)
                #             break
                #         else:
                #             elweights.append((el, f))
                #             elsum += f
                #     else:  # loop finished without break
                #         for (el, f) in elweights:
                #             ndr = [v*f/elsum for v in drange]
                #             for (at, targetel) in targetAtEls:
                #                 if el in at.disp_occ:
                #                     at.assignDisp(mode, ndr, el)
                #                 else:
                #                     logger.warning('DISPLACEMENTS file: '
                #                         'trying to address atom number '
                #                         +str(at.num)+' with wrong element. '
                #                         'Atom will be skipped.')
                #                     rp.setHaltingLevel(1)
        elif mode == 4:
            constraints.append((targetAtEls, ctype, value))
    sl.revert_unit_cell(uCellState)  # if modified by SYM_DELTA, go back
    # now read constraints
    for (targetAtEls, ctype, value) in constraints:
        if value.lower().startswith("link"):
            linkto = targetAtEls[0]
            for (at, el) in targetAtEls:  # [1:]:
                at.assignConstraint(ctype, targetel=el, linkAtEl=linkto)
        elif value.lower().startswith("ind("):
            try:
                ind = int(value.lower().split("ind(")[1].split(")")[0])
            except ValueError:
                logger.warning("DISPLACEMENTS file: Could not convert "+value
                               + " to index, skipping line.")
                continue
            for (at, el) in targetAtEls:
                at.assignConstraint(ctype, targetel=el, index=ind)
        else:
            if ctype == 3:
                occEl = value.split()[0]
                value = value.split(maxsplit=1)[-1]
            try:
                f = float(value.split()[0])
            except ValueError:
                logger.warning("DISPLACEMENTS file: Could not convert "+value
                               + " to float, skipping line.")
                continue
            for (at, el) in targetAtEls:
                if ctype == 3:
                    el = occEl
                at.assignConstraint(ctype, targetel=el, value=f)
    # Reading done, consistency check...
    o = ""
    if name:
        o = " "+name
    logger.debug("DISPLACEMENTS block{} was read successfully".format(o))
    for at in sl:
        if len(at.disp_occ) > 5:
            logger.error(
                'DISPLACEMENTS file: '+str(at)+' has occupations defined for '
                'more than 5 elements, which is not supported by TensErLEED.')
            raise RuntimeError("Too many elements")
        occlists = []
        error = False
        for k in at.disp_occ:
            occlists.append(at.disp_occ[k])
        for i in range(0, len(occlists[0])):
            totalocc = 0.
            for ol in occlists:
                if len(ol) <= i:
                    error = True
                    break
                else:
                    totalocc += ol[i]
            if error:
                logger.error('DISPLACEMENTS file: Lengths of occupation '
                             'lists per element differ for '+str(at))
                raise ValueError("Occupation lists per element must have same "
                                 "length.")
            if totalocc > 1 + 1e-4:    # some tolerance
                logger.error('DISPLACEMENTS file: Occupations for ' + str(at)
                             + ' sum to more than 1 for step '+str(i+1)+'.')
                raise ValueError("Occupations must sum to 1 or less.")
            elif totalocc < 1 - 1e-4 and len(at.disp_occ) > 4:
                logger.error(
                    'DISPLACEMENTS file: Occupations for ' + str(at)
                    + ' sum to less than 1 for step '+str(i+1)+', but a '
                    'vacancy cannot be added since there are already '
                    '5 elements.')
                raise ValueError("Occupations must sum to exactly 1 if five "
                                 "elements are already defined.")
    return deltas_required



"""ViPErLEED utilities for converting beam files to EXPBEAMS.csv."""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2024-03-27'
__license__ = 'GPLv3+'

from viperleed.utilities.beams.cli import BeamsUtilCLI

BeamsUtilCLI.run_as_script()



"""=================
    ViPErLEED
=================
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-03'
__license__ = 'GPLv3+'
__version__ = '0.12.1'


GLOBALS = {
    'USE_GUI': None,
    'version': __version__,
    'version_message': ('ViPErLEED (Vienna Package for Erlangen LEED) '
                        f'v{__version__}')
    }
# Name of environment variable specifying the path to the tensor-LEED
# source code. Typically the viperleed-tensorleed repository.
VIPERLEED_TENSORLEED_ENV = 'VIPERLEED_TENSORLEED'



"""
========================================
   ViPErLEED Graphical User Interface
========================================
 *** Module guilib.widgetdecorators ***


Author: Michele Riva
Created: 2020-01-29

This module provides decorators that can be used to decorate QWidget subclasses
"""

import wrapt
import PyQt5.QtCore as qtc
import PyQt5.QtGui as qtg
import PyQt5.QtWidgets as qtw

# from viperleed.guilib.decorators import ensure_decorates_class
# from viperleed.guilib.widgetslib import get_all_children_widgets
# from guilib.decorators import ensure_decorates_class
# from guilib.widgetslib import get_all_children_widgets
from viperleed import guilib as gl

mouse_event_types = (qtc.QEvent.MouseButtonPress, qtc.QEvent.MouseButtonRelease,
                     qtc.QEvent.MouseButtonDblClick,
                     qtc.QEvent.MouseMove, qtc.QEvent.Wheel)

@gl.ensure_decorates_class(qtw.QWidget)
@wrapt.decorator
def receive_mouse_broadcast(cls, _, args, kwargs):
    """
    Decorator for QWidget classes used for declaring that the class wants to
    receive mouse events even if they are not originally targeted to the
    instance. Instances of the decorated class will receive mouse events
    triggered on the ancestor or on any descendant (child, grandchild,
    grand-grandchild, etc.) of the ancestor, provided the ancestor does
    @broadcast_mouse events.
    """
    setattr(cls, '_wants_broadcast', True)

    return cls(*args, **kwargs)

@gl.ensure_decorates_class(qtw.QWidget)
@wrapt.decorator
def broadcast_mouse(cls, _, args, kwargs):
    """
    This is a class decorator intended to be used on a top-level window class.
    The decorated class will broadcast any mouse event to all children (as well 
    as grand-children, grand-grand-children, etc..) widgets that are instances
    of a QWidget subclass decorated with @receive_mouse_broadcast.
    (A specific mouse event will not be sent twice to the target of the event.)

    After broadcasting is complete, the mouse events will be handled as
    implemented in the class definition.
    """
    
    # Keep references to a few original method names that will be redefined, 
    # so that they can be called at appropriate places later on
    __orig_init = cls.__init__
    __orig_childEvent = cls.childEvent
    __orig_eventFilter = cls.eventFilter
    
    def __init__(self, *args, **kwargs):
        self.__broadcast_mouse_to = set()
        self.__last_broadcasted = dict()
        self.__polished_children = set()
        # All events generated within the instance will pass through its
        # eventFilter method defined below
        __orig_init(self, *args, **kwargs)
        
        self.installEventFilter(self)
    
    def childEvent(self, event):
        """
        Every time a child of type QWidget is added or removed, install/remove
        the eventFilter of this class on the child and all grand-children
        that want to @receive_mouse_broadcast.
        
        The original implementation of childEvent() of the decorated class
        is also called at the end.
        """
        # Notice that event.added() is fired at the very instant a widget is
        # added, but event.child() is not necessarily the right pointer to the 
        # child as the widget might not be polished yet
        if isinstance(event.child(), qtw.QWidget):
            children = gl.get_all_children_widgets(event.child())
            if (event.polished()
                    and event.child() not in self.__polished_children):
                self.__polished_children.update(children)
                self.__broadcast_mouse_to.update(child
                                                 for child in children
                                                 if hasattr(child,
                                                            '_wants_broadcast'))
                for child in children:
                    child.installEventFilter(self)
                    child.destroyed.connect(self.on_child_destroyed)
            elif event.removed() and event.child() in self.__polished_children:
                # Notice that event.child() will be in self.__polished_children
                # unless it is destroyed before the childEvent is processed
                # In that case the on_child_destroyed method takes care of
                # releasing the appropriate references
                self.__polished_children -= set(children)
                self.__broadcast_mouse_to -= set(children)
                for child in children:
                    child.removeEventFilter(self)
        # Create also a dictionary of "child: time-stamp of last event sent out"
        # that allows to not send multiple times the same event to the same
        # child, and prevents infinite recursion in case more than one child
        # wants to @receive_mouse_broadcast. Notice that the use of the sets
        # give a unique list of children here
        self.__last_broadcasted = {child: 0
                                   for child
                                   in self.__broadcast_mouse_to}
        # then call whatever implementation was there on the undecorated class
        # This should be a 'pass' except for QMainWindow and QSplitter
        __orig_childEvent(self, event)

    def on_child_destroyed(self, child):
        """
        This is the slot associated with the destroyed() signal of children
        widgets for which references are kept. The method releases these
        references so that the widgets cannot be accessed any longer after
        being destroyed.
        """
        self.__polished_children.discard(child)
        self.__broadcast_mouse_to.discard(child)

    def eventFilter(self, target, event):
        """
        Reimplementation of eventFilter that broadcasts all mouse events to 
        children widgets that want to @receive_mouse_broadcast.
        
        The original implementation of the decorated class is also always
        called.
        """
        # make sure that all child events that need to be addressed by the
        # childEvent method are correctly handled. This part is needed for
        # QMainWindow: if the centralWidget was already set (and consequently
        # already handled by childEvent above) when a new widget is added
        # as its child, no child event is emitted without the next lines
        if event.type() in (qtc.QEvent.ChildPolished, qtc.QEvent.ChildRemoved):
            self.childEvent(event)
        
        if event.type() in mouse_event_types:
            for child, timestamp in self.__last_broadcasted.items():
                if child != target and timestamp != event.timestamp():
                    # Notice the filtering is on the basis of event.timestamp()
                    # that has a 1 ms resolution: This should be good enough
                    # even for mouseMoveEvent, as the maximum polling
                    # rate of mice is 1kHz to date (Jan. 2020)
                    
                    if not child.underMouse():
                        edited_event = self.__edit_mouse_event(child, event)
                    else:
                        edited_event = event
                    if not self.__skip_event_broadcast(child, edited_event):
                        qtw.qApp.sendEvent(child, edited_event)
                self.__last_broadcasted[child] = event.timestamp()
        return __orig_eventFilter(self, target, event)

    def __edit_mouse_event(self, target, event):
        """
        Process the event by changing its location, so that widgets that are not
        under the mouse do not get animated (e.g., subclasses of QAbstractButton 
        like QPushButton animate parts of the widget)
        """
        top_left = target.rect().topLeft()
        new_local_pos = top_left - qtc.QPoint(5, 5)
        window_pos = target.mapTo(target.window(), new_local_pos)
        screen_pos = target.mapToGlobal(new_local_pos)
        new_source = qtc.Qt.MouseEventSynthesizedByApplication
        if event.type() == qtc.QEvent.Wheel:
            new_event = qtg.QWheelEvent(new_local_pos, screen_pos,
                                        event.pixelDelta(), event.angleDelta(),
                                        event.buttons(), event.modifiers(),
                                        event.phase(), event.inverted(),
                                        new_source)
        else:
            new_event = qtg.QMouseEvent(event.type(), new_local_pos, window_pos,
                                        screen_pos, event.button(),
                                        event.buttons(), event.modifiers(),
                                        new_source)
        new_event.setTimestamp(event.timestamp())
        new_event.setAccepted(event.isAccepted())
        return new_event

    def __skip_event_broadcast(self, target, event):
        """
        This function is used to treat specific events to specific widgets in
        a different way.

        Now implemented:
           * QLineEdit subclasses that do not haveFocus() should not get
             mouseMove nor mouseDoubleClick events as the Qt implementation does
             not check whether the widget hasFocus() nor whether it is
             underMouse() in the event handlers, so that QLineEdit respond
             anyway (in both cases a portion of the text gets highlighted as if
             it was selected)
        
        For future extension to other cases that I have not thought about,
        return a logic OR of the conditions
        """
        if (not target.isEnabled()
            or target.testAttribute(qtc.Qt.WA_TransparentForMouseEvents)):
            return True
        
        skip_line_edit = (isinstance(target, qtw.QLineEdit)
                          and not target.hasFocus()
                          and event.type() in (qtc.QEvent.MouseButtonDblClick,
                                               qtc.QEvent.MouseMove))
        return skip_line_edit


    setattr(cls, '__init__', __init__)
    setattr(cls, 'childEvent', childEvent)
    setattr(cls, 'eventFilter', eventFilter)
    setattr(cls, 'on_child_destroyed', on_child_destroyed)
    setattr(cls, '__edit_mouse_event', __edit_mouse_event)
    setattr(cls, '__skip_event_broadcast', __skip_event_broadcast)

    return cls(*args, **kwargs)



"""Module symmetry_eps of viperleed.calc.classes.rparams.special.

Defines the SymmetryEps class, a float with optional z value.
"""

__authors__ = (
    'Alexander Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-12-11'
__license__ = 'GPLv3+'

from functools import total_ordering
from numbers import Real

from .base import SpecialParameter


def _make_arithmetic(operation, allow_eps):
    """Return a direct and reverse version of operation.

    Parameters
    ----------
    operation : callable
        Arithmetic operation to be used for the float value and the
        z value of a SymmetryEps object. The methods returned are
        (essentially) operation(eps, other) and operation(other, eps),
        applied to both the main and z values of eps.
    allow_eps : bool
        Whether the second operand may be a SymmetryEps.

    Returns
    -------
    _direct : callable
        Returns operation(eps, other)
    _reverse : callable
        Returns operation(other, eps)
    """
    def _direct(self, other):
        if not allow_eps and isinstance(other, SymmetryEps):
            return NotImplemented
        if not isinstance(other, SymmetryEps):
            try:
                other = SymmetryEps(other)
            except (TypeError, ValueError):
                return NotImplemented
        value = operation(float(self), float(other))
        z_result = (operation(self.z, other.z) if self.has_z or other.has_z
                    else None)
        try:
            return self.__class__(value, z=z_result)
        except ValueError as exc:
            raise ValueError(exc) from None

    def _reverse(self, other):
        return _direct(self, other)
    return _direct, _reverse


@total_ordering
class SymmetryEps(float, SpecialParameter, param='SYMMETRY_EPS'):
    """SymmetryEps acts like a float but has an optional .z value.

    Used for interpreting the parameter SYMMETRY_EPS. The user
    can specify a second float value to be used for symmetry
    comparisons in the z direction (i.e., orthogonal to the
    surface). If no second value is given, the first value is
    used by default.
    """

    def __new__(cls, value, z=None):
        """Initialize instance."""
        cls._check_float_value(value)
        if z is not None:
            z = cls._check_float_value(z, 'z ')
        instance = super().__new__(cls, value)
        setattr(instance, '_z', z)
        return instance

    @staticmethod
    def _check_float_value(value, extra_msg=''):
        """Return a float version of value. Raise if not acceptable."""
        try:
            float_v = float(value)
        except (ValueError, TypeError):
            raise TypeError(f'SymmetryEps {extra_msg}value '
                            'must be float') from None
        if float_v <= 0:
            raise ValueError(f'SymmetryEps {extra_msg}value must be positive')
        return float_v

    @property
    def has_z(self):
        """Return whether self has a z value defined."""
        # About the disable: the member exists, it's created in __new__
        return self._z is not None  # pylint: disable=no-member

    @property
    def z(self):  # pylint: disable=invalid-name
        """Return the z value of this SymmetryEps."""
        # About the disable: the member exists, it's created in __new__
        if not self.has_z:
            return float(self)
        return self._z  # pylint: disable=no-member

    def __repr__(self):
        """Return a representation string for self."""
        txt = f'SymmetryEps({float(self)}'
        if self.has_z:
            txt += f', z={self.z}'
        return txt + ')'

    def __eq__(self, other):
        """Return self == other."""
        if not isinstance(other, (SymmetryEps, Real)):
            return NotImplemented
        was_real = not isinstance(other, SymmetryEps)
        if was_real:
            other = SymmetryEps(other)
        _eq = (float(self), self.z) == (float(other), other.z)
        if was_real:
            # Do not transform a False into a NotImplemented to
            # prevent python from falling back onto trying the
            # reverse operation other.__eq__(self) that returns
            # the incorrect result
            return _eq
        return _eq or NotImplemented

    def __lt__(self, other):
        """Return self < other."""
        if not isinstance(other, (SymmetryEps, Real)):
            return NotImplemented
        was_real = not isinstance(other, SymmetryEps)
        if was_real:
            other = SymmetryEps(other)
        _lt = float(self) < float(other) and self.z < other.z
        if was_real:
            # Do not transform a False into a NotImplemented to
            # prevent python from falling back onto trying the
            # reverse operation other.__gt__(self) that returns
            # the incorrect result
            return _lt
        return _lt or NotImplemented

    def __ne__(self, other):
        """Return self != other."""
        # Notice that, while normally python should take care
        # correctly of the default __ne__ implementation, we
        # want to make sure that we do not fall back onto
        # checking other.__ne__(self) when comparing to a
        # simple Real value. Otherwise we get the wrong result.
        # Reimplementing __ne__ is generally suggested when
        # inheriting from a built-in objects like here (float).
        # See github.com/python/cpython/issues/48645
        _eq = self.__eq__(other)
        if _eq is NotImplemented:
            return _eq
        return not _eq

    def __hash__(self):
        """Return hash(self)."""
        has_no_z = not self.has_z or self.z == float(self)
        if has_no_z:
            return super().__hash__()
        return hash((float(self), self.z))

    def __neg__(self):
        """Raise TypeError, as negative SymmetryEps are not valid."""
        raise TypeError('Negative SymmetryEps values are unsupported')

    def __pos__(self):
        """Return a copy of this SymmetryEps."""
        cls = self.__class__
        if self.has_z:
            return cls(float(self), z=self.z)
        return cls(float(self))

    def __round__(self, ndigits=None):
        """Return a rounded version of this SymmetryEps."""
        cls = self.__class__
        value = round(float(self), ndigits)
        if not self.has_z:
            return cls(value)
        return cls(value, round(self.z, ndigits))

    def __rpow__(self, other, modulo=None):
        """Raise TypeError not to allow other**self."""
        raise TypeError('unsupported operand type(s) for pow: '
                        f'{type(other).__name__!r} and '
                        f'{type(self).__name__!r}')

    def __pow__(self, other, modulo=None):
        """Return self**other."""
        cls = self.__class__
        if modulo is not None:
            # pow() supports the modulo argument only
            # when all three arguments are integers
            raise TypeError('3rd argument of pow() unsupported for '
                            f'{cls.__name__} objects')
        value = float(self)**other
        if not self.has_z:
            return cls(value)
        return cls(value, z=self.z**other)

    __add__, __radd__ = _make_arithmetic(float.__add__, allow_eps=True)
    __divmod__, _ = _make_arithmetic(float.__divmod__, allow_eps=False)
    __rdivmod__, _ = _make_arithmetic(float.__rdivmod__, allow_eps=False)
    __floordiv__, _ = _make_arithmetic(float.__floordiv__, allow_eps=False)
    __rfloordiv__, _ = _make_arithmetic(float.__rfloordiv__, allow_eps=False)
    __mod__, _ = _make_arithmetic(float.__mod__, allow_eps=False)
    __rmod__, _ = _make_arithmetic(float.__rmod__, allow_eps=False)
    __mul__, __rmul__ = _make_arithmetic(float.__mul__, allow_eps=False)
    __truediv__, _ = _make_arithmetic(float.__truediv__, allow_eps=False)
    __rtruediv__, _ = _make_arithmetic(float.__rtruediv__, allow_eps=False)
    __sub__, _ = _make_arithmetic(float.__sub__, allow_eps=True)
    __rsub__, _ = _make_arithmetic(float.__rsub__, allow_eps=True)



"""Module leedparameters of viperleed.guilib.leedsim.

========================================
   ViPErLEED Graphical User Interface
========================================
*** module guilib.leedsim.leedparameters ***

Created: 2021-01-23
Author: Michele Riva

This module contains the definition of the LEEDParameters class and
LEEDParametersList, that fully represent a viperleed.LEEDPattern and
viperleed.LEEDPatternList.
"""

from collections.abc import MutableMapping, MutableSequence
from configparser import ConfigParser
# import copy
import ast
from warnings import warn as warning   # eventually will replace with logging

import numpy as np

from viperleed import guilib as gl


class LEEDParameters(MutableMapping):
    """Dictionary-like container of LEED pattern parameters.

    Dictionary-like container of parameters needed to represent a
    viperleed.LEEDPattern. Keys cannot be deleted, and only acceptable
    keys can be set.
    """

    mandatory_keys = {'eMax': None,
                      'surfBasis': None,
                      'SUPERLATTICE': None,
                      'surfGroup': None,
                      'bulkGroup': None}

    # optional keys with their defaults:
    #     'name' will be the empty string unless it is defined in the
    #     dictionary used for instantiation, or if the LEEDParameters
    #     is loaded from a file (in which case it can be defined either
    #     as a section header or as a key of the configuration file;
    #     the latter takes precedence)
    optional_keys = {'screenAperture': 110,
                     'bulk3Dsym': None,
                     'beamIncidence': (0, 0),  # (theta, phi)
                     'name': '', }

    # keys that are not given as input, but are calculated at instantiation
    calculated_keys = {'bulkReciprocalBasis': None, }

    def __init__(self, data):
        # self.__data is the underlying dictionary that is accessed
        # when issuing self[key]
        if isinstance(data, LEEDParameters):
            # Just copy data into self.__data and skip all the rest
            self.__data = dict(data)
            return
        if not isinstance(data, (dict, ConfigParser)):
            raise TypeError("LEEDParameters: 'dict' or 'ConfigParser' needed. "
                            f"Found {type(data).__name__!r} instead")

        self.__data = {**self.mandatory_keys,
                       **self.optional_keys,
                       **self.calculated_keys}
        if isinstance(data, dict):
            self.__from_dict(data)
        else:
            self.__from_file(data)

        # Calculate the parameters in self.calculated_keys
        self._calculate_missing()

        # And fix the 3D symmetry operations of the bulk group by embedding
        # them into the PlaneGroup instance directly
        self._3d_ops_to_bulk_group()

    def __len__(self):
        """len(self)."""
        return len(self.__data)

    def __iter__(self):
        """Return iterator of self."""
        return iter(self.__data)

    def __setitem__(self, k, value):
        """Item setter.

        Parameters
        ----------
        k : str
            Name of parameter to set. Case-insensitive. Only
            acceptable parameters will be set.
        value : depends on k
            Value of the parameter to set.

        Raises
        ------
        Warning
            If k is not an acceptable parameter name.
        """
        if not isinstance(k, str):
            raise TypeError("LEEDParameters: keys need to be strings")
        try:
            idx = [key.lower() for key in self.__data].index(k.lower())
        except ValueError:
            warning(f"Unknown LEED parameter {k!r} will be ignored.")
        else:
            # key is there, but may be in the wrong case. Get the right one.
            k = tuple(self.__data.keys())[idx]
            data = {k: value}
            self._check_and_process_all(data)
            self.__data[k] = data[k]

    def __delitem__(self, k):
        """Not possible to delete items."""
        raise NotImplementedError

    def __getitem__(self, k):
        """Get item with key k."""
        return self.__data[k]

    def __contains__(self, k):
        """Return whether k is in self."""
        return k in self.__data

    def __repr__(self):
        """Return string representation of self."""
        return 'LEEDParameters(' + repr(self.__data).replace('\n', '') + ')'

    def __eq__(self, other):
        """Equality method for LEEDParameters.

        Instances are considered equal if they produce
        the very same LEED pattern, i.e., they have:
        - the same beamIncidence
        - the same surface group
        - the same bulk basis, expressed in the same coordinate
          system, but accounting for possible sign changes of
          one of the unit vectors
        - the same surface lattices, i.e., the same domains.
          Also in this case, sign changes are acceptable
          TODO: probably this thing with the sign change needs revisiting.

        The algorithm used assumes that both self and other
        would produce lattices with highest symmetry

        Returns
        -------
        True if LEEDParameters are equal, NotImplemented otherwise
        """
        if not isinstance(other, LEEDParameters):
            # Try to see if other can be used as an argument to the
            # LEEDParameters constructor
            try:
                other = LEEDParameters(other)
            except (TypeError, NameError, ValueError):
                # If instantiation raises one of the known exceptions
                return NotImplemented

        # (1) first a basic check
        if self is other:
            return True

        # (2) beamIncidence
        if self['beamIncidence'] != other['beamIncidence']:
            return NotImplemented

        # (3) surface plane group
        if self['surfGroup'] != other['surfGroup']:
            return NotImplemented

        # (4) reciprocal bulk bases, accounting for sign changes
        bulk_transform = np.dot(np.linalg.inv(other['bulkReciprocalBasis']),
                                self['bulkReciprocalBasis']).T
        if not np.allclose(np.abs(bulk_transform), gl.PlaneGroup.E, atol=1e-4):
            return NotImplemented

        # (5) check whether the real-space surface bases are the same,
        #     accounting for sign changes
        surf_transform = np.dot(other['surfBasis'],
                                np.linalg.inv(self['surfBasis']))
        if np.allclose(np.abs(surf_transform), gl.PlaneGroup.E):
            # (5.1) in this case, for the two patterns to be the same,
            #       we need also the bulk groups to be equal, including
            #       the operations that generate domains
            if ((self['bulkGroup'] == other['bulkGroup'])
                and (self['bulkGroup'].same_operations(other['bulkGroup'],
                                                       include_3d=True))):
                return True
            return NotImplemented

        # At this point we have:
        # * same beamIncidence
        # * same bulk basis (except sign changes)
        # * same surface group
        # * somewhat different surface bases

        # (6) The surface lattices could still be two symmetry-related
        #     domains. If this is the case, the two superlattice
        #     matrices should be related to one another by one of the
        #     symmetry operations of the bulk group
        # First we have to take into account that the BULK basis of
        # the other instance may have some sign change. This requires
        # transforming its superlattice to a common basis:
        super_other = np.dot(other['SUPERLATTICE'], bulk_transform.round())
        inv_super_self = np.linalg.inv(self['SUPERLATTICE'])

        # get the matrices that transform the superlattice of self
        # into the superlattice of other. Here we also have to
        # account for sign changes of the SURFACE basis
        sign_changes = (gl.PlaneGroup.E,   # no sign change
                        gl.PlaneGroup.C2,  # both vectors change sign
                        gl.PlaneGroup.Mx,  # b changes sign
                        gl.PlaneGroup.My)  # a changes sign
        # super_transforms is a "list" where the i-th element equals
        # np.dot(inv_super_self, np.dot(sign_changes[i], super_other))
        super_transforms = np.einsum("ilm,jl,mk->ijk",
                                     sign_changes,
                                     inv_super_self,
                                     super_other)

        # (6.1) if the transform is a bulk operation, it certainly
        #       needs to have only integer entries. Keep only the
        #       ones that do:
        int_transforms = [t.round().astype(int)
                          for t in super_transforms
                          if np.allclose(t, t.round())]
        if len(int_transforms) == 0:
            return NotImplemented

        # (6.2) check if indeed one of the transforms equals one
        #       of the bulk symmetry operations
        bulk_ops = self['bulkGroup'].operations(include_3d=True)
        for transform in int_transforms:
            transform = gl.two_by_two_array_to_tuple(transform)
            if transform in bulk_ops:
                return True
        return NotImplemented

    def __from_dict(self, data):
        """Create LEEDParameters from a dictionary.

        Checks that the dictionary contains the mandatory keys, and
        sets the optional ones to their default values, issuing some
        warnings if the warn keyword is True

        Parameters
        ----------
        data : dict or LEEDParameters
            Dictionary from which to create self.

        Returns
        -------
        None.
        """
        # Make all keys lowercase so that duplicates are removed,
        # and such that we don't modify the original data passed
        data = {k.lower(): v for k, v in data.items()}
        self._check_mandatory(data)
        self._check_and_process_all(data)
        for k, value in data.items():
            # Find the key in self that matches the one in data when
            # compared case-insensitive.  There should be at most
            # one.  All those that do not appear are neglected.
            keys = [kk for kk in self if kk.lower() == k.lower()]
            if len(keys):
                self[keys[0]] = value
            else:
                warning(f"Unknown LEED parameter {k!r} will be ignored.")

    def __from_file(self, parser):
        """Create LEEDParameters from a configuration file.

        Parameters
        ----------
        parser : ConfigParser
            The parser for the configuration file from which to
            initialize. It should contain only one section.

        Returns
        -------
        None.
        """
        if len(parser.sections()) > 1:
            raise TypeError("LEEDParameters: multiple structures found. Use "
                            "LEEDParametersList instead")
        # Take the section name as a 'name' attribute
        section = parser.sections()[0]
        self.__from_dict({**{'name': section}, **parser[section]})

    def _check_mandatory(self, data):
        """Check that all necessary keys are present in the dictionary."""
        if any(key not in data for key in self.mandatory_keys):
            # Perhaps they are there, just with the wrong case
            low = [key.lower() for key in data]
            missing = ', '.join(key for key in self.mandatory_keys
                                if key.lower() not in low)
            if missing:
                raise NameError("Not enough LEED parameters. "
                                f"Missing: {missing}")

    def _check_and_process_all(self, data):
        """Process data and check that the input is OK.

        Parameters
        ----------
        data : dict or LEEDParameters
            dictionary of parameters. Modified in place.

        Returns
        -------
        None.
        """
        self._process_input_data(data)
        self._check_acceptable(data)

    @staticmethod
    def _process_input_data(data):
        """Process data to correct format.

        This is called after _check_mandatory has been already
        executed to get the data dictionary with the correct types.

        Parameters
        ----------
        data : dict or LEEDParameters
            The data to be processed to get the correct types.
            Modified in place.

        Returns
        -------
        None.
        """
        for k, value in data.items():
            # Process all the values that should not be strings
            if k.lower() in ('emax', 'screenaperture'):
                try:
                    data[k] = float(value)
                except ValueError as err:
                    raise ValueError(f"LEEDParameters: invalid {k} "
                                     f"{value!r} found") from err

            elif k.lower() in ('surfbasis', 'superlattice'):
                dtype = float if k.lower() == 'surfbasis' else int
                try:
                    data[k] = LEEDParameters.__array_from_input(value, dtype)
                except ValueError as err:
                    raise ValueError(f"{err}: {k}") from err

            elif k.lower() in ('surfgroup', 'bulkgroup'):
                data[k] = gl.PlaneGroup(value)

            elif k.lower() == 'beamincidence':
                data[k] = LEEDParameters.__beam_incidence_from_input(value)

    @staticmethod
    def __array_from_input(input_matrix, dtype):
        """Convert input to 2x2 numpy.array.

        Parameters
        ----------
        input_matrix : str or iterable
        dtype : dtype
            Data type to use for the conversion from string

        Returns
        -------
        numpy.ndarray
            shape (2, 2)

        Raises
        ------
        ValueError
            if not possible to convert to (2, 2) array
        """
        matrix = None
        if isinstance(input_matrix, str):
            matrix = gl.string_matrix_to_numpy(input_matrix, dtype=dtype,
                                               needs_shape=(2, 2))
        else:
            try:
                matrix = np.asarray(input_matrix)
            except TypeError as err:
                raise ValueError("LEEDParameters: invalid "
                                 "matrix input") from err
        if matrix is None or matrix.shape != (2, 2):
            raise ValueError("LEEDParameters: invalid matrix shape")
        return matrix

    @staticmethod
    def __beam_incidence_from_input(angles):
        """Get beamIncidence tuple from input.

        Parameters
        ----------
        angles : iterable or str
            When an iterable, it should have only two elements.
            When a str, it should be possible to eval-it to a
            2-elements iterable.

        Raises
        ------
        ValueError
            If not possible to convert to beamIncidence angles

        Returns
        -------
        theta, phi
            Theta in [0, 90] range, phi in [0,360).
        """
        if isinstance(angles, str):
            angles = ast.literal_eval(angles)
        if (not isinstance(angles, (list, tuple, np.ndarray))
                or np.shape(angles) != (2,)
                or not -90 <= angles[0] <= 90):
            raise ValueError("LEEDParameters: invalid beamIncidence "
                             f"entry found: {angles}")
        return gl.conventional_angles(*angles)

    @staticmethod
    def _check_acceptable(data_dict):
        """Check that the dict input is acceptable.

        Parameters
        ----------
        data_dict : dict or LEEDParameters

        Raises
        ------
        ValueError if not data_dict contains unacceptable data.
        """
        for key, value in data_dict.items():
            if key.lower() == 'emax' and value <= 0:
                raise ValueError("Maximum LEED energy should be positive.")
            if (key.lower() == 'screenaperture'
                    and (value <= 0 or value > 180)):
                raise ValueError("screenAperture should be between "
                                 f"0 and 180. Found {value} instead.")
            if (key.lower() in ('surfbasis', 'superlattice')
                    and np.shape(value) != (2, 2)):
                raise ValueError("Lattice basis and SUPERLATTICE need to have "
                                 f"a (2, 2) shape. Found {np.shape(value)} "
                                 "instead")
            if (key.lower() == 'superlattice'
                    and any(np.abs(np.asarray(value) % 1).ravel() > 1e-4)):
                raise ValueError("SUPERLATTICE needs to be an integer-valued "
                                 "matrix")

    def _calculate_missing(self):
        """Calculate extra parameters that used only internally.

        Calculates the extra parameters that are not in the original
        input but are useful to keep
        """
        # (1) bulkReciprocalBasis
        # build a dummy real-space surface lattice just to get its reciprocal
        # basis, from which one can find the bulk reciprocal basis. Also,
        # round a bit the result (8 decimal digits).
        bulk_basis = np.dot(self['SUPERLATTICE'].T,
                            gl.Lattice(self['surfBasis']).reciprocal_basis)
        self['bulkReciprocalBasis'] = bulk_basis.round(decimals=8)

    def _3d_ops_to_bulk_group(self):
        """Edit the 'bulkGroup' entry to contain screws/glides.

        Edit the 'bulkGroup' entry, which should be a PlaneGroup
        instance, such that it contains the correct glide and screw
        symmetry operations.
        """
        # do something only if it makes sense
        if self['bulk3Dsym'] is None:
            return

        # get dummy bulk to get the lattice shape
        dummy_bulk = gl.Lattice(self['bulkReciprocalBasis'],
                                space='reciprocal',
                                group=self['bulkGroup'])
        bulk_shape = dummy_bulk.cell_shape

        # And apply the bulk operations to the bulk group
        self['bulkGroup'].screws_glides = (self['bulk3Dsym'], bulk_shape)


class LEEDParametersList(MutableSequence):
    """1-D only list of LEEDParameters.

    1-D only list of LEEDParameters that are suitable for producing
    LEED patterns, i.e., only structures with the same bulk lattice
    and symmetry can be added after instantiation.

    During instantiation checks that all are consistent.
    """

    def __init__(self, data=None, keep_duplicates=False):
        """Initialize LEEDParametersList.

        Parameters
        ----------
        data : iterable or LEEDParametersList
            Each element is either dict, ConfigParser, or
            LEEDParameters. Each element will be checked for
            compatibility, i.e., all should have the same bulk.
        keep_duplicates : bool (default=False)
            If False, only one of the entries in data that
            would produce identical LEED patterns (i.e., same
            spots, same symmetry domains, same symmetry relations
            between spots) is kept.

        Returns
        -------
        None.
        """
        super().__init__()
        # self.__list is the list that underlies this container,
        # i.e., the one that is accessed when calling self[index].
        # It needs initializing to empty because it is accessed
        # in _process_input_data
        self.__list = []
        self.__list = self._process_input_data(data, keep_duplicates)

    def __delitem__(self, index):
        """Remove item."""
        del self.__list[index]

    def __getitem__(self, index):
        """Get item at index."""
        # Do this type-preserving. If a slice, should
        # return same type, if index, just an element
        if isinstance(index, slice):
            return LEEDParametersList(self.__list[index])
        return self.__list[index]

    def __setitem__(self, idx, data):
        """Set item at idx to data.

        The item is set only if it is an acceptable LEEDParameter,
        given the data already present in self.

        Parameters
        ----------
        idx : int or slice
        data : dict, ConfigParser or LEEDParameters
        """
        data = self._to_leed_parameters(data)
        self._check_acceptable(*data)
        if not data:
            return
        self.__list[idx] = data

    def __len__(self):
        """Return len(self)."""
        return len(self.__list)

    def __repr__(self):
        """Return string representiation of self."""
        txt = 'LEEDParametersList([\n'
        txt += '\n'.join(repr(d) for d in self)
        txt += '\n])'
        return txt

    def insert(self, idx, data, keep_duplicates=False):
        """Insert data at index.

        The data is inserted only if it is compatible with the
        data already present in self.

        Parameters
        ----------
        idx : int
            Index at which to insert
        data : dict, ConfigParser, or LEEDParameters
            The data to be inserted. Will be checked for compatibility.
        keep_duplicates : bool (default=False)
            If False, data is inserted only if it is not a duplicate
            of elements already present in self.
        """
        data = self._process_input_data(data, keep_duplicates)
        # when inserting, we should make sure that the data is actually
        # a single LEEDParameters to avoid creating a list with entries
        # of different type at different indices
        if not data:
            # Do nothing if trying to insert nothing
            return
        if len(data) > 1:
            # More than one parameter passed.
            # Not acceptable for insertion.
            raise ValueError("Cannot insert multiple LEEDParameters at one "
                             "index. Use LEEDParametersList.extend(params) "
                             "or a list of LEEDParametersList")
        self.__list.insert(idx, data[0])

    def append(self, data, keep_duplicates=False):
        """Insert a single element at the end.

        The data is appended only if it is compatible with the
        data already present in self.

        Parameters
        ----------
        data : dict, ConfigParser, or LEEDParameters
            The data to be inserted. Will be checked for compatibility.
        keep_duplicates : bool (default=False)
            If False, data is appended only if it is not a duplicate
            of elements already present in self.
        """
        self.insert(len(self), data, keep_duplicates)

    def extend(self, data, keep_duplicates=False):
        """Insert a elements from an iterable at the end.

        Only those entries that are compatible with the
        data already present in self are appended.

        Parameters
        ----------
        data : iterable or LEEDParametersList
            Each element is either dict, ConfigParser, or LEEDParameters
            The data to be inserted. Each element will be checked for
            compatibility.
        keep_duplicates : bool (default=False)
            If False, append only those elements in data that are not
            a duplicate of elements already present in self.
        """
        for datum in data:
            self.append(datum, keep_duplicates)

    def set_beam_incidence(self, theta=None, phi=None):
        """Set the 'beamIncidence' key to the given angles.

        The same angles will be set for all the entries in self.

        Parameters
        ----------
        theta : float, default=None
            Polar incidence angle. Should be in the [-90, 90] range.
            Normal icidence is theta=0. No modification if not given
            or if None.
        phi : float
            Azimuthal incidence angle. Positive counterclockwise,
            measured with respect to the positive x axis of the
            Cartesian coordinates in which the real-space bases
            are expressed. Will be taken modulo 360.
        """
        if theta is None and phi is None:
            return
        old_angles = self[0]['beamIncidence']
        if theta is None:
            theta = old_angles[0]
        elif phi is None:
            phi = old_angles[1]
        try:
            theta = float(theta)
            phi = float(phi)
        except TypeError as err:
            raise TypeError("Invalid angles. "
                            "Expected numbers.") from err
        if (theta, phi) == old_angles:
            return
        new_angles = gl.conventional_angles(theta, phi)
        if new_angles == old_angles:
            return
        for param in self:
            param['beamIncidence'] = new_angles

    def _process_input_data(self, data, keep_duplicates=False):
        """Prcess the given data.

        Do all the processing needed to accept the input data, i.e.,
        converting each entry to a LEEDParameters instance, verify
        that the data passed is consistent, and remove duplicates
        if needed.

        Parameters
        ----------
        data : iterable or LEEDParametersList
            Each element is either dict, ConfigParser, or
            LEEDParameters. Each element will be checked for
            compatibility.
        keep_duplicates : bool, default=False
            If False, discard those elements in data that are
            a duplicate of elements already present in self,
            as well as duplicates within data.
        """
        data = self._to_leed_parameters(data)
        self._check_acceptable(*data)
        if not keep_duplicates:
            # first remove duplicates from data only
            data = gl.remove_duplicates(data)
            # then keep only those that are not already in the list.
            # This way, if at earlier calls we wanted to keep
            # duplicates, they remain in self
            data = [d for d in data if d not in self]
        return data

    @staticmethod
    def _to_leed_parameters(data):
        """Convert data to a list of LEEDParameters.

        Parameters
        ----------
        data : dict, ConfigParser, LEEDParameters, or iterable of the same

        Returns
        -------
        list of LEEDParameters
        """
        if not data:
            ret = []
        elif isinstance(data, dict):
            # delegate to LEEDParameters the check on the dictionary
            ret = [LEEDParameters(data)]
        elif isinstance(data, ConfigParser):
            ret = [LEEDParameters({**{'name': section}, **data[section]})
                   for section in data]
        elif isinstance(data, LEEDParameters):
            ret = [data]
        elif isinstance(data, LEEDParametersList):
            ret = data
        elif isinstance(data, (list, tuple, np.ndarray)):
            # Here we check if any of the entries in the list
            # is a LEEDParametersList first, in which case we
            # replace the entry with as many LEEDParameters
            # instances as there are in the LEEDParametersList
            ret = []
            for elem in data:
                if isinstance(elem, LEEDParametersList):
                    ret.extend(elem)
                else:
                    ret.append(LEEDParameters(elem))
        else:
            raise TypeError("LEEDParametersList: invalid data type. Expected "
                            "dict, ConfigParser, LEEDParameters, "
                            "LEEDParametersList, or list, tuple, or "
                            "np.ndarray of the previous. "
                            f"Found {type(data).__name__!r} instead")
        return ret

    def _check_acceptable(self, *data):
        """Check that the data is acceptable.

        Data is considered acceptable if the bulk lattices have the
        same bases (includng possibly sign changes), and if the
        bulk groups contain the same symmetry operations.  At the
        same time, this method replaces all 'eMax' and 'screenAperture'
        keys with the maxima among all the data (and of self).

        Parameters
        ----------
        *data : dict, ConfigParser, or LEEDParameters

        Raises
        ------
        ValueError
            If any of the data is inconsistent

        Returns
        -------
        None.

        """
        # First store the largest eMax and screenAperture.
        # Will be set as the eMax and screenAperture values
        # if the data is acceptable
        all_params = (*self, *data)
        if len(all_params) == 0:
            return
        emax = max(param['eMax'] for param in all_params)
        aperture = max(param['screenAperture'] for param in all_params)

        # check consistency of bulk
        inv_bulk_basis = np.linalg.inv(all_params[0]['bulkReciprocalBasis'])
        bulk_group = all_params[0]['bulkGroup']
        for param in all_params[1:]:
            # Bulk lattices should be the same, except, possibly, for a
            # sign change. Compute the matrix that transforms one into
            # the other. Then check that the transform is, except for a
            # sign change, the identity matrix. Since the bulk bases
            # are the reciprocal ones, one would need to calculate
            #        T = ((B'*)^-1 @ B*)^t,
            # but it is more convenient to calculate
            #        T* = B'* @ B*^(-1)
            # since abs(T) == E <--> abs(T*) == E
            transform = np.dot(param['bulkReciprocalBasis'], inv_bulk_basis)
            if not np.allclose(np.abs(transform), gl.PlaneGroup.E):
                raise ValueError("LEEDParametersList: Inconsistent bulk bases "
                                 "found in the input parameters")

            # Bulk groups also should be the same, but excluding the
            # 3D symmetry as this only affects how manysymmetry
            # domains will be generated
            this_group = param['bulkGroup']
            same_bulk_group = (
                bulk_group == this_group
                and bulk_group.same_operations(this_group, include_3d=False)
                )
            if not same_bulk_group:
                raise ValueError("LEEDParametersList: Inconsistent symmetry "
                                 "operations of bulk lattices in the "
                                 "input parameters")

        # Check passed. Set eMax and screenAperture to the max
        for param in all_params:
            param['eMax'] = emax
            param['screenAperture'] = aperture



"""Section Rfactor."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-11'
__license__ = 'GPLv3+'

import logging
import os
from pathlib import Path
import shutil
import subprocess

import numpy as np

from viperleed.calc import DEFAULT_WORK
from viperleed.calc.files import iorfactor
from viperleed.calc.files import iotensors
from viperleed.calc.files.iorefcalc import readFdOut
from viperleed.calc.lib import leedbase
from viperleed.calc.lib.checksums import validate_multiple_files

try:
    from viperleed.calc.wrapped.rfactor import r_factor_new as rf
except ImportError:
    _HAS_NEW_R_FACTOR = False
else:
    from viperleed.calc.wrapped.error_codes import check_ierr
    _HAS_NEW_R_FACTOR = True

logger = logging.getLogger(__name__)


def rfactor(sl, rp, index, for_error=False, only_vary=None):                    # TODO: Parameters __doc__
    """Runs the r-factor calculation for either the reference calculation
    (index 11) or the superpos (index 12)."""
    if rp.THEO_ENERGIES.n_energies < 2:
        logger.info("Only one theoretical energy found: Cannot "
                    "calculate a meaningful R-Factor. Stopping...")
        return []
    if for_error and rp.best_v0r is None:
        # Cannot proceed with planned section
        err_msg = ("Cannot calculate R-factor for error without a stored"
                   "V0r value. Execute a normal R-factor calculation first.")
        rp.setHaltingLevel(3)
        logger.error(err_msg)
        raise RuntimeError(err_msg)

    # Both refcalc (11) and superpos should work fine with new R-factor
    if index == 11:
        name = "refcalc"
    else:
        name = "superpos"
    _need_spectra = (index == 11 and not len(rp.refcalc_fdout)
                     or index == 12 and not len(rp.superpos_specout))
    if _need_spectra:
        _fetch_and_check_spectra(rp, index, name)
    theobeams = rp.theobeams[name]
    expbeams = rp.expbeams

    # Branch off for new R factor calculation
    if not rp.R_FACTOR_LEGACY and _HAS_NEW_R_FACTOR:                            # TODO: we may want to raise some errors if we want to use it but it's not there
        return run_new_rfactor(sl, rp, for_error, name, theobeams, expbeams)
    return run_legacy_rfactor(sl, rp, for_error, name, theobeams, index, only_vary)


def _fetch_and_check_spectra(rp, index, name):
    """Look for full-dynamic spectra, then check consistency."""
    if index == 11:
        fn = Path("refcalc-fd.out")
    else:
        fn = Path("superpos-spec.out")
    directory = None
    path = None
    if fn.is_file():
        directory = DEFAULT_WORK
        path = fn
    elif ("OUT" / fn).is_file():
        directory = "OUT"
        path = "OUT" / fn
    elif index == 11:
        # try getting from Tensors
        iotensors.getTensors(rp.TENSOR_INDEX, required=False)
        directory = Path(f"Tensors_{rp.TENSOR_INDEX:03d}")
        if ("Tensors" / directory / fn).is_file():
            path = "Tensors" / directory / fn

    if path:
        logger.warning("R-factor calculation was called without stored "
                       f"spectrum data. Reading from file {fn} in {directory} "
                       "folder...")
    else:
        logger.error("Cannot execute R-factor calculation: no stored "
                     f"spectrum data and no {fn} file was found.")
        raise RuntimeError("No spectrum data found")
    try:
        theobeams, theospec = readFdOut(readfile=path)
    except Exception:
        logger.error(f"Failed to read {path}")
        raise

    # if we haven't returned, then it was read. check data vs ivbeams:
    if index == 11:
        rp.refcalc_fdout = theospec
    else:
        rp.superpos_specout = theospec
    rp.theobeams[name] = theobeams

    eps = 1e-3
    consistent = (len(rp.ivbeams) == len(theobeams)
                  and all(any(ivbeam.isEqual(theobeam, eps=eps)
                              for theobeam in theobeams)
                              for ivbeam in rp.ivbeams))
    if not consistent:
        logger.error("The list of beams read from IVBEAMS is not equivalent "
                     f"to the list of beams in {path}. R-Factor calculation "
                     "cannot proceed.")
        raise ValueError("Contradiction in beam sets")


def run_new_rfactor(sl, rp, for_error, name, theobeams, expbeams):
    logger.debug("Using new R-factor calculation. This is still experimental!")
    which_r = rp.R_FACTOR_TYPE

    if which_r == 1:
        n_derivs = 1
    elif which_r == 2:
        n_derivs = 0
    else:
        check_ierr(701, logger)

    (_, theo_range,
     iv_shift,
     intpol_step) = iorfactor.prepare_rfactor_energy_ranges(
        rp, theobeams, for_error,
        n_expand=(rp.INTPOL_DEG - 1) // 2
        )
    out_grid = np.arange(theo_range.min,
                         theo_range.max + 0.1 * intpol_step,
                         intpol_step)

    # find correspondence experimental to theoretical beams:
    beamcorr = leedbase.getBeamCorrespondence(sl, rp)
    # integer & fractional beams
    iorf = []
    for (i, beam) in enumerate(rp.expbeams):
        if beam.hk[0] % 1.0 != 0.0 or beam.hk[1] % 1.0 != 0.0:
            iorf.append(2)
        else:
            iorf.append(1)
    iorf.extend([0] * (len(rp.ivbeams) - len(rp.expbeams)))

    n_beams = len(beamcorr) - beamcorr.count(-1)  # -1 indicates no correspondence
    (exp_grid, exp_id_start,
     exp_n_E_beams,
     exp_intensities_in) = iorfactor.beamlist_to_array(rp.expbeams)
    corr_theobeams = []
    for i in beamcorr:
        if i != -1:
            corr_theobeams.append(theobeams[i])
    (theo_grid,
     theo_id_start,
     theo_n_E_beams,
     theo_intensities_in) = iorfactor.beamlist_to_array(corr_theobeams)

    deg = rp.INTPOL_DEG
    v0i = rp.V0_IMAG

    # Should we do reordering averaging etc. using the subroutine? - unused right now... # TODO
    averaging_scheme = np.int32(np.arange(n_beams) + 1)  # Fortran index

    # create buffer arrays
    (
        exp_e_start_beams_out,
        exp_n_e_beams_out,
        exp_intpol_int,
        exp_intpol_deriv_y,
        ierr,
    ) = rf.alloc_beams_arrays(n_beams, len(out_grid))
    check_ierr(ierr, logger)

    # interpolation and derivative
    ierr = rf.prepare_beams(
        e_grid_in=exp_grid,
        intensities=exp_intensities_in,
        e_start_beams=exp_id_start + 1,  # Fortran indices
        n_e_beams=exp_n_E_beams,
        deg=deg,
        n_derivs=n_derivs,
        e_grid_out=out_grid,
        e_start_beams_out=exp_e_start_beams_out,
        n_e_beams_out=exp_n_e_beams_out,
        intpol_intensity=exp_intpol_int,
        intpol_derivative=exp_intpol_deriv_y,
    )
    # check for erros
    check_ierr(ierr, logger)

    # Same for theoretical beams now
    # create buffer arrays
    (
        theo_e_start_beams_out,
        theo_n_e_beams_out,
        theo_intpol_int,
        theo_intpol_deriv_y,
        ierr,
    ) = rf.alloc_beams_arrays(n_beams, len(out_grid))
    check_ierr(ierr, logger)

    # interpolation and derivative
    ierr = rf.prepare_beams(
        e_grid_in=theo_grid,
        intensities=theo_intensities_in,
        e_start_beams=theo_id_start + 1,  # Fortran indices
        n_e_beams=theo_n_E_beams,
        deg=deg,
        n_derivs=n_derivs,
        e_grid_out=out_grid,
        e_start_beams_out=theo_e_start_beams_out,
        n_e_beams_out=theo_n_e_beams_out,
        intpol_intensity=theo_intpol_int,
        intpol_derivative=theo_intpol_deriv_y,
    )
    # check for erros
    check_ierr(ierr, logger)

    # For Factor R2 we use intesities for further calculations, otherwise the Y function
    if which_r == 1:
        # Make Y-functions
        rf.pendry_y_beamset(
            exp_intpol_int,
            exp_intpol_deriv_y,  # derivative is overwritten by the Y function
            exp_e_start_beams_out,
            exp_n_e_beams_out,
            v0i,
        )

        rf.pendry_y_beamset(
            theo_intpol_int,
            theo_intpol_deriv_y,  # derivative is overwritten by the Y function
            theo_e_start_beams_out,
            theo_n_e_beams_out,
            v0i,
        )

        exp_data = exp_intpol_deriv_y
        theo_data = theo_intpol_deriv_y

    elif which_r == 2:
        exp_data = exp_intpol_int
        theo_data = theo_intpol_int

        # optimize V0r and calculate R factor

        ## settings for V0r optimization
    bounds = iv_shift.min, iv_shift.max
    v0r_range = np.array(
        [round(bound / intpol_step) for bound in bounds], dtype="int32"
    )
    v0r_center = int((v0r_range[0] + v0r_range[1]) / 2)
    start_guess = np.array([v0r_range[0], v0r_center, v0r_range[1]], dtype="int32")

    # TODO: make below into User Parameters
    fast_search = False
    tol_r = (1 - 5e-2,)
    tol_r_2 = (1 - 5e-2,)
    max_fit_range = 6

    (
        best_v0r_step,
        best_v0r,
        best_rfac,
        _,  # n_v0r_evaluated
        R_beams,
        numerators,
        denominators,
        n_overlapp_beams,
        ierr,
    ) = rf.r_beamset_v0r_opt_on_grid(
        which_r,
        v0r_range,
        start_guess,
        fast_search,
        intpol_step,
        exp_data,
        theo_data,
        exp_e_start_beams_out,
        theo_e_start_beams_out,
        exp_n_e_beams_out,
        theo_n_e_beams_out,
        tol_r=tol_r,
        tol_r_2=tol_r_2,
        max_fit_range=max_fit_range,
    )

    check_ierr(ierr, logger)

    logger.info(
        "With inner potential shift of {:.2f} eV: "
        "R = {:.4f}\n".format(best_v0r, best_rfac)
    )
    rp.best_v0r = best_v0r

    # seperate into integer and fractional beams
    n_groups = 2
    grouping = np.int32(iorf)  # 1 is integer, 2 is fractional

    # Call grouping routine
    (r_factor_groups, _, ierr) = rf.r_beamtype_grouping(  # n_overlapp_groups unused
        which_r, R_beams, numerators, denominators, n_overlapp_beams, n_groups, grouping
    )
    # error 903 and 904 are acceptable - that just means there are no integer/fractional beams

    check_ierr(ierr, logger)

    (r_int, r_frac) = list(r_factor_groups)

    if np.isclose(best_rfac, 0):
        logger.error(
            "R-Factor reported as zero. This means "
            "something went wrong in the reference "
            "calculation or in the R-factor calculation."
        )
        rp.setHaltingLevel(2)
    elif best_rfac > 2:
        logger.error(
            f"R-Factor reported as {best_rfac}. This means "
            "something went wrong in the reference "
            "calculation or in the R-factor calculation."
        )
        rp.setHaltingLevel(2)
    else:
        rp.last_R = best_rfac
        rp.stored_R[name] = (best_rfac, r_int, r_frac)

    if rp.PLOT_IV["plot"]:
        # Plot R-factor
        outname = "Rfactor_plots_{}.pdf".format(name)
        aname = "Rfactor_analysis_{}.pdf".format(name)
        if rp.PLOT_IV["overbar"]:
            labelstyle = "overbar"
        else:
            labelstyle = "minus"
        # labelwidth = max([beam.getLabel(style=labelstyle)[1] for beam in rp.expbeams]) # TODO currently unused? why?

        labels = [beam.label for beam in expbeams]

        # Apply v0r shift to arrays for plotting

        shifted_intpol_exp = np.copy(exp_intpol_int)
        shifted_intpol_theo = np.copy(theo_intpol_int)
        shifted_y_exp = np.copy(exp_data)
        shifted_y_theo = np.copy(theo_data)
        shifted_E_start_exp = np.copy(exp_e_start_beams_out)
        shifted_E_start_theo = np.copy(theo_e_start_beams_out)
        shifted_n_exp = np.copy(exp_n_e_beams_out)
        shifted_n_theo = np.copy(theo_n_e_beams_out)

        """
            np.savetxt("viper_exp_start.csv", exp_e_start_beams_out, delimiter = ',')
            np.savetxt("viper_exp_n_beams.csv", exp_n_e_beams_out, delimiter = ',')
            np.savetxt("viper_exp_y.csv", exp_yfunc, delimiter = ',')
            np.savetxt("viper_exp_intensity.csv", exp_intpol_intensity, delimiter = ',')
            np.savetxt("viper_theo_start.csv", theo_e_start_beams_out, delimiter = ',')
            np.savetxt("viper_theo_n_beams.csv", theo_n_e_beams_out, delimiter = ',')
            np.savetxt("viper_theo_y.csv", theo_yfunc, delimiter = ',')
            np.savetxt("viper_theo_intensity.csv", theo_intpol_intensity, delimiter = ',')
            np.savetxt("viper_theo_energies.csv", theo_grid, delimiter = ',')
            np.savetxt("viper_exp_energies.csv", exp_grid, delimiter = ',')
            """

        ierr = rf.apply_beamset_shift(
            shifted_intpol_exp,
            shifted_E_start_exp,
            shifted_n_exp,
            shifted_intpol_theo,
            shifted_E_start_theo,
            shifted_n_theo,
            shift=best_v0r_step,
            fill_outside=1,
        )
        check_ierr(ierr, logger)
        ierr = rf.apply_beamset_shift(
            shifted_y_exp,
            np.copy(exp_e_start_beams_out),
            np.copy(exp_n_e_beams_out),
            shifted_y_theo,
            np.copy(theo_e_start_beams_out),
            np.copy(theo_n_e_beams_out),
            shift=best_v0r_step,
            fill_outside=1,
        )
        check_ierr(ierr, logger)

        # shifted_E_start_exp must be now same as shifted_E_start_theo -> no need to pass both
        # same for n_e_beams
        iorfactor.writeRfactorPdf_new(
            n_beams,
            labels,
            R_beams,
            out_grid,
            shifted_E_start_exp,
            shifted_n_exp,
            shifted_intpol_exp,
            shifted_intpol_theo,
            shifted_y_exp,
            shifted_y_theo,
            outName=outname,
            analysisFile=aname,
            v0i=rp.V0_IMAG,
            formatting=rp.PLOT_IV,
        )
    rfaclist = list(R_beams)
    return rfaclist


def run_legacy_rfactor(sl, rp, for_error, name, theobeams, index, only_vary):

    if index == 11:
        theospec = rp.refcalc_fdout
    elif index == 12:
        theospec = rp.superpos_specout
    # WEXPEL before PARAM, to make sure number of exp. beams is correct
    try:
        iorfactor.writeWEXPEL(sl, rp, theobeams, for_error=for_error)
    except Exception:
        logger.error("Exception during writeWEXPEL: ")
        raise
    try:
        iorfactor.writeRfactPARAM(rp, theobeams,
                                  for_error=for_error,
                                  only_vary=only_vary)
    except Exception:
        logger.error("Exception during writeRfactPARAM: ")
        raise

    # get fortran files and compile
    try:
        tl_source = rp.get_tenserleed_directory()
        tl_path = tl_source.path
        libpath = tl_path / "lib"
        libname = next(libpath.glob("rfacsb*"))                                 # StopIteration??
        srcpath = tl_path / "src"
        srcname = next(srcpath.glob("rfactor.*"))                               # StopIteration??
        shutil.copy2(libname, libname.name)  # Copy here from source
        shutil.copy2(srcname, srcname.name)  # Copy here from source
    except Exception:
        logger.error(
            "Error getting TensErLEED files for r-factor calculation:"
            )
        raise
    if rp.SUPPRESS_EXECUTION:
        logger.warning("SUPPRESS_EXECUTION parameter is on. R-factor "
                       "calculation will not proceed. Stopping...")
        rp.setHaltingLevel(3)
        return []
    # validate checksums
    if not rp.TL_IGNORE_CHECKSUM:
        files_to_check = (Path(libpath) / libname,
                          Path(srcpath) / srcname)
        validate_multiple_files(files_to_check, logger,
                                "R-factor", rp.TL_VERSION)

    logger.info("Compiling fortran input files...")
    rfacname = f"rfactor-{rp.timestamp}"
    if rp.FORTRAN_COMP[0] == "":
        rp.getFortranComp()
    # compile
    _pre, _post = rp.FORTRAN_COMP
    ctasks = [
        (f"{_pre} -o rfacsb.o -c", libname.name, _post),
        (f"{_pre} -o main.o -c", srcname.name, _post),
        (f"{_pre} -o {rfacname}", "rfacsb.o main.o", _post)
        ]

    # Compile
    compile_log = Path("compile-rfactor.log")
    try:
        leedbase.fortran_compile_batch(ctasks, logname=compile_log)
    except Exception:
        logger.error("Error compiling fortran files: ", exc_info=True)
        # move compile log to compile_logs directory
        leedbase.copy_compile_log(rp, compile_log, "rfactor-compile")
        raise

    # move log file to supp
    try:
        shutil.move(compile_log, "compile_logs" / compile_log)
    except OSError:
        logger.warning(f"Could not move {compile_log} to SUPP")
    # run
    rfaclogname = Path(rfacname).with_suffix(".log")
    logger.info(
        "Starting R-factor calculation...\n"
        f"R-factor log will be written to file {rfaclogname}"
        )
    try:
        with rfaclogname.open("w") as log:
            subprocess.run(
                Path(rfacname).resolve(),
                input=theospec,
                encoding="ascii",
                stdout=log,
                stderr=log,
                )
    except Exception:
        logger.error("Error during R-factor calculation. "
                     "Also check R-factor log file.")
        raise
    logger.info("Finished R-factor calculation. Processing files...")
    # rename and move files
    try:
        os.rename("WEXPEL", "rfactor-WEXPEL")
    except OSError:
        logger.warning("Failed to rename R-factor input "
                       "file WEXPEL to rfactor-WEXPEL")
    try:
        os.rename("PARAM", "rfactor-PARAM")
    except OSError:
        logger.warning("Failed to rename R-factor input "
                       "file PARAM to rfactor-PARAM")
    if not Path("ROUT").is_file():
        logger.error("No ROUT file was found after R-Factor calculation!")
        rp.setHaltingLevel(2)
        return []

    # read output
    if for_error:
        try:
            rfaclist = iorfactor.readROUTSHORT()
        except Exception:
            logger.error("Error reading ROUTSHORT file",
                         exc_info=rp.is_debug_mode)
            rp.setHaltingLevel(2)
        return rfaclist

    try:
        (rfac, r_int, r_frac), v0rshift, rfaclist = iorfactor.readROUT()
    except Exception:  # TODO catch correct exception
        logger.error("Error reading ROUT file", exc_info=rp.is_debug_mode)
        rp.setHaltingLevel(2)
        return []

    logger.info("With inner potential shift of {:.2f} eV: "
                "R = {:.4f}\n".format(v0rshift, rfac))
    rp.best_v0r = v0rshift
    dir_list = [Path(), Path("OUT")]
    for dir_name in dir_list:
        for f_name in dir_name.glob(f"R_OUT_{rp.timestamp}*"):
            if not f_name.is_file():
                continue
            try:  # delete old R_OUT files
                f_name.unlink()
            except Exception:
                pass
    if rfac <= 0.00001:
        logger.warning(
            "ROUT reports R-Factor as zero. This may mean "
            "something went wrong in the reference "
            "calculation or in the R-factor calculation. "
            "If you are comparing with pseudo-experiment data, "
            "you can ignore this warning."
            )

    f_name = f"R_OUT_{name}_{rp.timestamp}_R={rfac:.4f}"
    rp.last_R = rfac
    rp.stored_R[name] = (rfac, r_int, r_frac)
    try:
        os.rename("ROUT", f_name)
    except OSError:
        logger.warning("Failed to rename R-factor "
                       f"output file ROUT to {f_name}")
    if len(rfaclist) != len(rp.expbeams):
        logger.warning("Failed to read R-Factors per "
                       "beam from R-factor output file ROUT.")
        rfaclist = [-1] * len(rp.expbeams)

    if rp.PLOT_IV["plot"]:
        outname = f"Rfactor_plots_{name}.pdf"
        aname = f"Rfactor_analysis_{name}.pdf"
        labelstyle = "overbar" if rp.PLOT_IV["overbar"] else "minus"
        labelwidth = max(beam.getLabel(style=labelstyle)[1]
                         for beam in rp.expbeams)
        try:
            iorfactor.writeRfactorPdf(
                [(b.getLabel(lwidth=labelwidth, style=labelstyle)[0], r)
                 for b, r in zip(rp.expbeams, rfaclist)],
                outName=outname,
                analysisFile=aname,
                v0i=rp.V0_IMAG,
                formatting=rp.PLOT_IV,
                )
        except Exception:                                                       # TODO catch correct exception
            logger.warning("Error plotting R-factors.", exc_info=True)
    return rfaclist
